#include "GlobalHeap.h"

#define SER 0

GlobalHeap::GlobalHeap()
{
  freeList = NULL;
  pthread_mutex_init(&globalMutex, NULL);
}

GlobalHeap::~GlobalHeap()
{
}

/* init():
 * initialize global heap
 * immediatelly reserve 40KB of memory
 */
int GlobalHeap::init()
{
  freeList = NULL;
  pthread_mutex_init(&globalMutex, NULL);
  //immediatally request 40KB
  void *initialMemory = malloc(10 * SUPERBLOCK_SIZE);
  free(initialMemory);
  return 0;
}

/* GlobalHeap malloc
 * returns a pointer to the beginning of a block 
 * of size = muliple of SUPERBLOCK_SIZE
 * sets the first 4 bytes = size the block
 */
void* GlobalHeap::malloc(size_t size)
{
  assert(size > 0);
  //round up to the closest power of 2
  size = ROUNDUP(size, SUPERBLOCK_SIZE) * SUPERBLOCK_SIZE;
  //# of superblocks needed for this request
  int n_superblocks = size / SUPERBLOCK_SIZE;
  void *ptr = NULL;

  pthread_mutex_lock(&globalMutex);

  //try to find that many contiguous superblocks in the free list
  ptr = getConsecutiveSuperBlocks(n_superblocks);
  //need to expand the heap
  if (ptr == NULL) 
    {
      ptr = mem_sbrk(size);
    }

  assert(ptr != NULL);
  setBlockSize(ptr, size);  

  pthread_mutex_unlock(&globalMutex);

  assert((char *)ptr >= mem_heap_lo());
  assert((char *)ptr + size - 1 <= mem_heap_hi());

  return ptr;
}

/* Global heap free
 * returns a block of size of a muliple of SUPERBLOCK_SIZE
 * to the sorted list of free superblocks
 * if the block is big, splits into multiple superblocks
 * ptr must point to the very beginning of the entire superblock
 */
void GlobalHeap::free(void *ptr)
{
  assert(ptr != NULL);
 
  size_t size = getBlockSize(ptr);

  assert(size >= SUPERBLOCK_SIZE);
  assert(size % SUPERBLOCK_SIZE == 0);

  //# of superblocks corresponding to this block to free
  int n_superblocks = size / SUPERBLOCK_SIZE;
  for (int i = 0; i < n_superblocks; i++)
    {
      setBlockSize((char *)ptr + i * SUPERBLOCK_SIZE, SUPERBLOCK_SIZE);
      setBlockNext((char *)ptr + i * SUPERBLOCK_SIZE, (char *)ptr + (i + 1) * SUPERBLOCK_SIZE);
      if (i == n_superblocks - 1) //last one
	setBlockNext((char *)ptr + i * SUPERBLOCK_SIZE, NULL);
    }

  //return [ptr, last in the block] to the free list
  pthread_mutex_lock(&globalMutex);
  returnMultipleSuperBlocks(ptr, (char *)ptr + (n_superblocks - 1) * SUPERBLOCK_SIZE);
  pthread_mutex_unlock(&globalMutex);
}

/* getConsecutiveSuperBlock
 * returns a poitner to the first superblock
 * of a contiguous n_superblocks in memory
 * or NULL if not found
 */
void* GlobalHeap::getConsecutiveSuperBlocks(int n_superblocks)
{
  void *p = freeList; //current pointer in list
  void *prev = NULL; //previous to p
  void *last = NULL; //last contigous block in p
  void *ret = NULL; 

  while (p != NULL)
    {
      //check if p, next(p), next(next(p)), ... [n_superblocks total]
      //are consecutive blocks    
      last = p;
      int i;
      for (i = 1; i < n_superblocks; i++)
	{
	  //check if the next block start right after "last"
	  if (getBlockNext(last) != ((char*)last + SUPERBLOCK_SIZE))
	    break;
	  assert(getBlockNext(last) >= (char*)last + SUPERBLOCK_SIZE);
	  assert((char *)last + SUPERBLOCK_SIZE - 1 < mem_heap_hi());
	  last = getBlockNext(last);
	}
      
      //success: found n contiguous superblocks
      if (i == n_superblocks)
	{
	  //unlink from the list
	  if (prev) 
	    {
	      setBlockNext(prev, getBlockNext(last));
	    }
	  //if removing from the head of the list, update head
	  if (freeList == p)
	    {
	      freeList = getBlockNext(last);
	    }

	  ret = p;
	  break;
	}

      //not found; advance current pointer p
      prev = last;
      assert(getBlockNext(last) == NULL || (getBlockNext(last) > last && getBlockNext(last) > p));
      p = getBlockNext(last);
    }

  assert(freeList ==  NULL || ret == NULL || !(freeList >= ret && freeList < (char*)last + SUPERBLOCK_SIZE));
  return ret;
}

/* returnMultipleSuperBlocks
 * returns a linked list of contiguous superblocks
 * first -> next(first) -> ... -> last
 * into a sorted freeList
 */
void GlobalHeap::returnMultipleSuperBlocks(void *first, void *last)
{
  assert(first != NULL);
  assert(last != NULL);
  assert(first <= last);
  assert(getBlockNext(last) == NULL);

  void *next = freeList;
  void *prev = NULL;
  while (next && next < first)
    {
      prev = next;
      next = getBlockNext(next);
    }

  assert(prev == NULL || prev < first);
  assert(next == NULL || next > last);
  assert(next == NULL || next < mem_heap_hi());

  setBlockNext(last, next);
  if (prev == NULL)
    {
      freeList = first;
    }
  else
    {
      setBlockNext(prev, first);
    }

  assert(last < mem_heap_hi());
  assert(getBlockNext(last) == NULL || getBlockNext(last) < mem_heap_hi());
  assert(freeList < mem_heap_hi());
}

/* check()
 * returns 0 if the heap is consistent
 * return -1 otherwise
 */
int GlobalHeap::check()
{
  pthread_mutex_lock(&globalMutex);
  void *p = freeList;
  int ret = 0;
  while (p && ret == 0)
    {
      //the free list is sorted by pointers' addresses
      //so the next element must be >= cur pointer + superblock size
      //this also guarantees no loops etc. in the linked list
      if (getBlockNext(p) != NULL && getBlockNext(p) < (char *)p + SUPERBLOCK_SIZE)
	{
	  ret = -1;
	}
      
      //beginning of block in heap?
      if (p < mem_heap_lo())
	{
	  ret = -1;
	}

      //end of block in heap?
      if ((char *)p + SUPERBLOCK_SIZE - 1 > mem_heap_hi())
	{
	  ret = -1;
	}
      
      //all superblocks in free list must be of size SUPERBLOCK_SIZE
      if (getBlockSize(p) != SUPERBLOCK_SIZE)
	{
	  ret = -1;
	}

      p = getBlockNext(p);
    }
  pthread_mutex_unlock(&globalMutex);
  return ret;
}
