#include "LocalHeap.h"

#define SER 0

LocalHeap::LocalHeap()
{
  pthread_mutex_init(&localMutex, NULL);
  for (int i = 0; i <= MAX_BLOCK_POW; i++)
    buckets[i] = NULL;
}

LocalHeap::~LocalHeap()
{
}

/* init():
 * initializes the local heap
 */
int LocalHeap::init(GlobalHeap *globalHeap, int heapId)
{
  pthread_mutex_init(&localMutex, NULL);
  this->globalHeap = globalHeap;
  this->heapId = heapId;
  for (int i = 0; i <= MAX_BLOCK_POW; i++)
    buckets[i] = NULL;
  return 0;
}

/* malloc():
 * returns a pointer to a block of size >= size 
 * if local heap runs out free blocks of that size
 * a new superblock is allocated from the global heap
 */
void *LocalHeap::malloc(size_t size)
{
  int blockIndex = getBlockIndex(size);
  assert(blockIndex <= MAX_BLOCK_POW);
  assert(blockIndex >= 0);
  
  pthread_mutex_lock(&localMutex);
  
  //find a superblock of given size with at least 1 free block
  void *block = findEmptyBlock(blockIndex);
  //no such block, need to request a new superblock from the global heap
  if (block == NULL)
    block = requestNewSuperBlock(blockIndex);
  assert(block != NULL);

  pthread_mutex_unlock(&localMutex);

  assert(block >= mem_heap_lo());
  assert((char*)block + size - 1 <= mem_heap_hi());
  return block;
}

/* free():
 * returns a block pointed by ptr to local heap
 * if there are multiple free superblocks in corresponding bucket
 * returns all but 1 free superblocks to the global heap
 */
void LocalHeap::free(void *ptr)
{
  void *superblock = getSuperBlockFromPointer(ptr, mem_heap_lo());
  assert(superblock != NULL);
  size_t size = getBlockSize(superblock);
  int blockIndex = getBlockIndex(size);

  pthread_mutex_lock(&localMutex);

  _freeBlock(superblock, ptr, blockIndex);
  
  //superblock used to be full, now has 1 block free
  //need to add back to the linked list
  if (getBlockUsage(superblock) == getBitmaskSize(blockIndex) - 1)
    {
      _addToList(superblock, blockIndex);
    }

  //if the superblock becomes empty and it's not the only
  //superblock in this bucket, return to the global heap
  bool empty = false;
  if (isSuperBlockEmpty(superblock, blockIndex) && 
      !(buckets[blockIndex] == superblock && getBlockNext(superblock) == NULL))
    {
      _removeFromList(superblock, blockIndex);
      empty = true;
    }

  pthread_mutex_unlock(&localMutex);

  //return a superblock to global heap, if any
  if (empty)
    {
      setBlockSize(superblock, SUPERBLOCK_SIZE);
      globalHeap->free(superblock);
    }
}

/* requestNewSuperBlock():
 * allocates a superblock block and puts it in the right bracket
 * in this local heap
 */
void *LocalHeap::requestNewSuperBlock(int blockIndex)
{
  void *superblock = globalHeap->malloc(SUPERBLOCK_SIZE);
  assert(superblock);

  //convert global superblock into a superblock struct of local heap
  setBlockSize(superblock, 1 << blockIndex);
  setBlockNext(superblock, NULL);
  setBlockOwner(superblock, heapId);
  setBlockUsage(superblock, 0);
  _clearBitmask(superblock, blockIndex);

  _addToList(superblock, blockIndex);
  //choose the first block 
  return _markBlock(superblock, 0, blockIndex);
}

/* _clearBitmask
 * zeros-out the entire bitmask of superblock
 */
void LocalHeap::_clearBitmask(void *superblock, int blockIndex)
{
  unsigned int *bitmask = (unsigned int *)getBlockBitmask(superblock);
  assert(bitmask != NULL);
  int nbits = (SUPERBLOCK_SIZE - LOCAL_DATA_START) / (1 << blockIndex);
  for (int i = 0; i < nbits; i += 32)
    {
      *bitmask = 0;
      bitmask++;
    }
  setBlockUsage(superblock, 0);
}

/* _removeFromList
 * removes superblock from the linked list buckets[blockIndex]
 */
void LocalHeap::_removeFromList(void *superblock, int blockIndex)
{
  void *p = buckets[blockIndex];
  void *prev = NULL;
  while (p != superblock)
    {
      prev = p;
      p = getBlockNext(p);
    }
  assert(p == superblock);

  //remove from head
  if (prev == NULL)
    {
      buckets[blockIndex] = getBlockNext(p);
    }
  else
    {
      setBlockNext(prev, getBlockNext(p));
    }
}

/* _addToList
 * appends (at the front) superblock to the linked list buckets[blockIndex]
 */
void LocalHeap::_addToList(void *superblock, int blockIndex)
{
  setBlockNext(superblock, buckets[blockIndex]);
  buckets[blockIndex] = superblock;
}

/* isSuperBlockEmpty
 * returns true if the *entire* bitmask in superblock is empty
 * false otherwise
 */
bool LocalHeap::isSuperBlockEmpty(void *superblock, int blockIndex)
{
  return getBlockUsage(superblock) == 0;
}

/* findEmptyBlock
 * returns a pointer to the first free block of size 2^blockIndex
 * or NULL if no such block exist
 */
void *LocalHeap::findEmptyBlock(int blockIndex)
{
  void *ret = NULL;
  int nbits = getBitmaskSize(blockIndex);
  assert(nbits > 0);

  //linked list keeps only non-full superblocks
  //this is guaranteed to find an empty block
  void *superblock = buckets[blockIndex];
  if (superblock != NULL)
    {
      assert(getBlockUsage(superblock) < getBitmaskSize(blockIndex));

      //pointer to the beginning of the bitmask
      unsigned int *p = (unsigned int *)getBlockBitmask(superblock);
      
      for (int i = 0; i + 32 < nbits; i += 32)
	{
	  //if there's at least 1 free block
	  if (*p != (unsigned int)-1)
	    {
	      ret = _markBlock(superblock, i / 32, blockIndex);
	      break;
	    }
	  p++;
	}
      //not found in full bytes; must be in the remainder [last byte]
      if (ret == NULL)
	{
	  ret = _markBlock(superblock, nbits / 32, blockIndex, nbits % 32);
	}

      //just made the superblock full, remove from the linked list
      if (getBlockUsage(superblock) == getBitmaskSize(blockIndex))
	{
	  _removeFromList(superblock, blockIndex);
	}
    }

  return ret;
}

/* _markBlock
 * marks and returns the first free block in superblock 
 * incremenets superblock's usage
 * maxm - optional; tries only the first maxm bits of the kbyte-th byte
 * returns a pointer to a block
 * should never return NULL
 */
void *LocalHeap::_markBlock(void *superblock, int kbyte, int blockIndex, int maxm)
{
  assert(maxm <= 32);
  assert(getBlockUsage(superblock) < getBitmaskSize(blockIndex));

  void *ret = NULL;
  unsigned int *bitmaskByte = (unsigned int *)getBlockBitmask(superblock) + kbyte;
  int index = findIndexOfFirstZeroBit(*(int *)bitmaskByte);

  assert(index <= 31);
  assert(index >= 0);

  //mark the bit
  *bitmaskByte |= ((unsigned int)1) << index;
  //corresponding block
  ret = (char *)getBlockData(superblock) + (kbyte * 32 + index) * (1 << blockIndex);
  
  setBlockUsage(superblock, getBlockUsage(superblock) + 1);

  assert(getBlockUsage(superblock) > 0);
  assert(ret != NULL);
  return ret;
}

/* _freeBlock
 * marks the bit corresponding to ptr as unmarked
 * decrements superblock's usage
 * in superblock 
 */
void LocalHeap::_freeBlock(void *superblock, void *ptr, int blockIndex)
{
  char *bitmask = (char *)getBlockBitmask(superblock);
  //distance of ptr from the beginning of the data block in this superblock
  size_t len = (char *)ptr - (char *)getBlockData(superblock);
  int index = len / (1 << blockIndex);
  
  bitmask += index / 8;
  *bitmask ^= 1 << (index % 8);

  //decrement block's usage
  setBlockUsage(superblock, getBlockUsage(superblock) - 1);

  assert(getBlockUsage(superblock) < getBitmaskSize(blockIndex));
}

/* check():
 * returns 0 if the local heap is consistent
 * return -1 otherwise
 */
int LocalHeap::check()
{
  pthread_mutex_lock(&localMutex);

  int ret = 0;
  //for each bucket size
  for (int bucket_index = 0; bucket_index <= MAX_BLOCK_POW && ret == 0; bucket_index++)
    {
      //check each superblock on its linked list
      void *p = buckets[bucket_index];
      while (p != NULL && ret == 0)
	{
	  //superblock's size should be equal to the bucket size
	  if (getBlockSize(p) != (1 << bucket_index))
	    {
	      ret = -1;
	    }
	  
	  //beginning of superblock out of heap?
	  if (p < mem_heap_lo())
	    {
	      ret = -1;
	    }
	  
	  //end of block out of heap?
	  if ((char *)p + SUPERBLOCK_SIZE - 1 > mem_heap_hi())
	    {
	      ret = -1;
	    }

	  //does this superblock belong to this local heap?
	  if (getBlockOwner(p) != heapId)
	    {
	      ret = -1;
	    }

	  //usage more than maximum bitmask size?
	  if (getBlockUsage(p) > getBitmaskSize(bucket_index))
	    {
	      ret = -1;
	    }
	  
	  //make sure the number of 1's in the bitmask is the same as block's usage
	  int nOnes = 0;
	  for (int i = 0; i < getBitmaskSize(bucket_index); i += 8)
	    {
	      char *bitmaskByte = (char *)getBlockBitmask(p) + i/8;
	      for (int j = 0; j < 8; j++)
		if (((*bitmaskByte) >> j)&1)
		  {
		    nOnes++;
		  }
	    }
	  if (nOnes != getBlockUsage(p))
	    {
	      ret = -1;
	    }

	  //loop
	  if (p == getBlockNext(p))
	    {
	      ret = -1;
	    }

	  //go to the next superblock on this bucket's list
	  p = getBlockNext(p);
	}
      
    }

  pthread_mutex_unlock(&localMutex);
  return ret;
}
