#include <cstdio>
#include "allocator_util.h"

/* ROUNDUP():
 * returns ceil(a/b)
 */
int ROUNDUP(int a, int b)
{
  if ((a/b)*b < a) return a/b + 1;
  return a/b;
}

/* isBigRequest():
 * returns true if size corresponds to a big request
 * returns false otherwise
 */
bool isBigRequest(size_t size)
{
  return size > (1 << MAX_BLOCK_POW);
}

/* getBlockIndex():
 * returns a index of a block that 'size' belongs to
 * -1 if 'size' out of maximum block size
 */
int getBlockIndex(size_t size)
{
  for (int i = 0; i <= MAX_BLOCK_POW; i++)
    if ( (1<<i) >= size)
      return i;
  return -1;
}

/* getSuperBlockFromPointer():
 * returns a pointer to the superblock
 * containing block pointed by 'ptr'
 */
void *getSuperBlockFromPointer(void *ptr, void *mem_lo)
{
  unsigned long long a = (unsigned long long)((char*)ptr - (char*)mem_lo);
  a %= SUPERBLOCK_SIZE;
  return (void *)((char *)ptr - a);
}

/* getBitmaskSize():
 * returns the exact number of blocks at blockIndex size in a superblock
 * = # of useful bits in a bitmask
 */
int getBitmaskSize(int blockIndex)
{
  return (SUPERBLOCK_SIZE - LOCAL_DATA_START) / (1 << blockIndex);
}

/* setBlockSize():
 * sets (super)block size to val
 */
void setBlockSize(void *ptr, size_t val)
{
  *(size_t*)ptr = val;
}

/* getBlockSize():
 * returns size of superblock pointed by ptr
 */
size_t getBlockSize(void *ptr)
{
  return *(size_t *)ptr;
}

/* getBlockNext():
 * return a pointer to the superblock following
 * the superblock pointed by 'ptr' in a linked list
 */
void* getBlockNext(void *ptr)
{
  return *(void **)((char*)ptr + LOCAL_NEXT_START);
}

/* setBlockNext():
 * sets the next pointer of a superblock 'ptr' to 'val'
 */
void setBlockNext(void *ptr, void *val)
{
  ptr = (char *)ptr + LOCAL_NEXT_START;
  *(void **)ptr = val;
}

/* getBlockOwner():
 * returns id of the heap that holds superblock pointed by 'ptr'
 */
size_t getBlockOwner(void *ptr)
{
  return *(size_t*)((char*)ptr + LOCAL_OWNER_START);
}

/* setBlockOwner():
 * sets superblock's owner to 'owner'
 */
void setBlockOwner(void *ptr, size_t owner)
{
  *(size_t*)((char *)ptr + LOCAL_OWNER_START) = owner;
}

/* getBlockUsage():
 * returns the number of used blocks in a superblock
 */
size_t getBlockUsage(void *ptr)
{
  return *(size_t*)((char *)ptr + LOCAL_USAGE_START);
}

/* setBlockUsage
 * sets the number of used blocks in a superblock to 'usage'
 */
void setBlockUsage(void *ptr, size_t usage)
{
  *(size_t*)((char *)ptr + LOCAL_USAGE_START) = usage;
}

/* getBlockBitmask():
 * returns pointer to a bitmask of superblock pointed by 'ptr'
 */
void* getBlockBitmask(void *ptr)
{
  return (void *)((char*)ptr + LOCAL_BITMASK_START);
}

/* getBlockData():
 * return pointer to the first block in superblock 'ptr'
 */
void *getBlockData(void *ptr)
{
  return (void *)((char*)ptr + LOCAL_DATA_START);
}

/* findIndexOfFirstZeroBit():
 * returns zero-based index of the first zero bit in  val
 * [least significant zero]
 * return 32 if all 32 bits are 1
 */
int findIndexOfFirstZeroBit(int val)
{
  val = ~val;
  return doBitCount((val&(-val))-1);
}

/* doBitCount():
 * counts the number of 1s in an integer
 */
int doBitCount(unsigned int u)
{
  unsigned int uCount;

  uCount = u
    - ((u >> 1) & 033333333333)
    - ((u >> 2) & 011111111111);
  return
    ((uCount + (uCount >> 3))
     & 030707070707) % 63;
}

